#include"server_UDP.h"

void init_struct_UDP_client(UDP_client *client_list, unsigned int nNmax)
{
    client_list->Nmax = nNmax;
    client_list->NbClient = 0;
    client_list->id_client = (unsigned int *)calloc(nNmax, sizeof(unsigned int));
    client_list->sin_client = (SOCKADDR_IN *)calloc(nNmax, sizeof(SOCKADDR_IN));
}

void reset_UDP_client(UDP_client *client_list)
{
    client_list->NbClient = 0;
    for (int i=0 ; i<client_list->Nmax ; i++)
    {
        client_list->id_client[i] = 0;
        //client_list->sin_client[i] = {0,0,0,0};
    }
}

void delete_struct_UDP_client(UDP_client *client_list)
{
    if (client_list->id_client != NULL)
    {
        free(client_list->id_client);
        client_list->id_client = NULL;
    }
    if (client_list->sin_client != NULL)
    {
        free(client_list->sin_client);
        client_list->sin_client = NULL;
    }
}

void add_UDP_client(SOCKET *server_socket, UDP_client *client_list)
{
    socklen_t sin_len = sizeof(SOCKADDR_IN);
    char trig = '0';

    if (recvfrom(*server_socket, &trig, 1, MSG_WAITALL, (SOCKADDR *)(&client_list->sin_client[client_list->NbClient]), &sin_len) < 0)
    {
        printf("error using recvfrom in add_UDP_client...");
    }
    client_list->id_client[client_list->NbClient] = client_list->NbClient;
    client_list->NbClient++;
}

void clear_UDP_client(UDP_client *client_list, unsigned int client_ID)
{
    int i = 0;
    UDP_client temp_struct = {};
    init_struct_UDP_client(&temp_struct, client_list->Nmax);
    temp_struct.NbClient = client_list->NbClient-1;
    {
        for (i=0 ; i<client_ID ; i++)
        {
            temp_struct.id_client[i] = i;
            temp_struct.sin_client[i] = client_list->sin_client[i];
        }
        
        if (client_ID == (client_list->NbClient - 1))
        {
            // if the cleared client is the last, the copy is already done
        }
        else // copy from id+1 to nb client
        {
            for (i=client_ID+1 ; i<client_list->NbClient ; i++)
            {
                temp_struct.id_client[i-1] = i-1;
                temp_struct.sin_client[i-1] = client_list->sin_client[i];
            }
        }
    }
    //recopy temporary structure into client list
    reset_UDP_client(client_list);
    for (i=0 ; i<temp_struct.NbClient ; i++)
    {
        client_list->id_client[i] = i;
        client_list->sin_client[i] = temp_struct.sin_client[i];
    }
    client_list->NbClient = temp_struct.NbClient;
    delete_struct_UDP_client(&temp_struct);
}

void *UDP_line_routine(void *p_data)
{
    UDP_server_info *l_UDP_server = (UDP_server_info *)p_data;
    while (l_UDP_server->client_list->NbClient < l_UDP_server->client_list->Nmax)
    {
        printf("Waiting for new client connecting to socket 7600\n");
        add_UDP_client(l_UDP_server->server_socket, l_UDP_server->client_list); 
        printf("new client connected, total number of client: %i\n", l_UDP_server->client_list->NbClient);
    }

    return NULL;
}

void launch_UDP_line_server(UDP_server_info *sinfo, pthread_t *npthread)
{
    if (pthread_create(npthread, NULL, UDP_line_routine, sinfo)!=0)
    {
        printf("pthread_create error\n");
    }
}

void send_UDP_line(UDP_server_info *sinfo, void *buffer, int buffer_size, int client_ID)
{
    if ((sinfo->client_list->NbClient > 0) && (client_ID < (int)(sinfo->client_list->NbClient)))
    {
        if (client_ID >= 0)
        {
            sendto(*sinfo->server_socket, buffer, buffer_size, MSG_DONTWAIT, (SOCKADDR *)(&sinfo->client_list->sin_client[client_ID]), sizeof(SOCKADDR_IN));
        }
        else
        {
            for (int i=0 ; i<sinfo->client_list->NbClient ; i++)
            {
                sendto(*sinfo->server_socket, buffer, buffer_size, MSG_DONTWAIT, (SOCKADDR *)(&sinfo->client_list->sin_client[i]), sizeof(SOCKADDR_IN));
            }
        }
    }
}

void *UDP_image_routine(void *p_data)
{
    UDP_server_info *l_UDP_server = (UDP_server_info *)p_data;
    while (l_UDP_server->client_list->NbClient < l_UDP_server->client_list->Nmax)
    {
        printf("Waiting for new client connecting to socket 7610\n");
        add_UDP_client(l_UDP_server->server_socket, l_UDP_server->client_list); 
        printf("new client connected, total number of client: %i\n", l_UDP_server->client_list->NbClient);
    }

    return NULL;
}

void launch_UDP_image_server(UDP_server_info *sinfo, pthread_t *npthread)
{
    if (pthread_create(npthread, NULL, UDP_image_routine, sinfo)!=0)
    {
        printf("pthread_create error\n");
    }
}

void send_UDP_image(UDP_server_info *sinfo, void *buffer, int buffer_size, int client_ID)
{
    if ((sinfo->client_list->NbClient > 0) && (client_ID < (int)(sinfo->client_list->NbClient)))
    {
        if (client_ID >= 0)
        {
            sendto(*sinfo->server_socket, buffer, buffer_size, MSG_DONTWAIT, (SOCKADDR *)(&sinfo->client_list->sin_client[client_ID]), sizeof(SOCKADDR_IN));
        }
        else
        {
            for (int i=0 ; i<sinfo->client_list->NbClient ; i++)
            {
                sendto(*sinfo->server_socket, buffer, buffer_size, MSG_DONTWAIT, (SOCKADDR *)(&sinfo->client_list->sin_client[i]), sizeof(SOCKADDR_IN));
            }
        }
    }
}
